// SPDX-License-Identifier: UNLICENSED
pragma solidity ^0.8.28;

import {IUniswapV2Pair} from "./IUniswapV2Pair.sol";
import {IERC20} from "./IERC20.sol";
import {IUniswapV2Callee} from "./IUniswapV2Callee.sol";


contract ArbitrageManager is IUniswapV2Callee {
    uint256 constant f = 997;

    function sqrt(uint256 x)
        public
        pure
        returns (uint128)
    {
        if (x == 0) return 0;
        else{
            uint256 xx = x;
            uint256 r = 1;
            if (xx >= 0x100000000000000000000000000000000) { xx >>= 128; r <<= 64; }
            if (xx >= 0x10000000000000000) { xx >>= 64; r <<= 32; }
            if (xx >= 0x100000000) { xx >>= 32; r <<= 16; }
            if (xx >= 0x10000) { xx >>= 16; r <<= 8; }
            if (xx >= 0x100) { xx >>= 8; r <<= 4; }
            if (xx >= 0x10) { xx >>= 4; r <<= 2; }
            if (xx >= 0x8) { r <<= 1; }
            r = (r + x / r) >> 1;
            r = (r + x / r) >> 1;
            r = (r + x / r) >> 1;
            r = (r + x / r) >> 1;
            r = (r + x / r) >> 1;
            r = (r + x / r) >> 1;
            r = (r + x / r) >> 1;
            uint256 r1 = x / r;
            return uint128 (r < r1 ? r : r1);
        }
    }
    
    function getAmountOut(uint256 amountIn, uint256 reserveIn, uint256 reserveOut)
        public
        pure
        returns (uint256 amountOut)
    {
        require(amountIn > 0, "UniswapV2Library: INSUFFICIENT_INPUT_AMOUNT");
        require(reserveIn > 0 && reserveOut > 0, "UniswapV2Library: INSUFFICIENT_LIQUIDITY");
        uint256 amountInWithFee = amountIn * 997;
        uint256 numerator = amountInWithFee * reserveOut;
        uint256 denominator = reserveIn * 1000 + amountInWithFee;
        amountOut = numerator / denominator;
    }

    function optimalIn(uint256 X_A, uint256 Y_A, uint256 X_B, uint256 Y_B)
        public
        pure
        returns(uint256)
    {
        uint256 k = f * X_B / 1000 + f ** 2 / 1000 * X_A / 1000;
        uint256 phi = sqrt(f ** 2 * X_A * Y_B * X_B * Y_A / 1000**2);
        uint256 psi = Y_A * X_B;
        
        if (psi >= phi) return 0;
        else return (phi - psi) / k;
    }
    
  function flashArbitrage(address firstPair, address secondPair, bool tokenDir)
        public
        returns (uint256 gain)
    {
        IUniswapV2Pair pairA = IUniswapV2Pair(firstPair);
        IUniswapV2Pair pairB = IUniswapV2Pair(secondPair);

        IERC20 firstToken = tokenDir ? IERC20(pairA.token0()) : IERC20(pairA.token1());
        IERC20 secondToken = tokenDir ? IERC20(pairA.token1()) : IERC20(pairA.token0());
        
        (uint256 X_A, uint256 Y_A,) = pairA.getReserves();
        (uint256 X_B, uint256 Y_B,) = pairB.getReserves();

        uint256 amountIn = optimalIn(tokenDir ? X_B : X_A, tokenDir ? Y_B : Y_A, tokenDir ? X_A : X_B, tokenDir ? Y_A : Y_B); 
        uint256 firstAmountOut = getAmountOut(amountIn, tokenDir ? X_A : Y_A, tokenDir ? Y_A : X_A);
        uint256 secondAmountOut = getAmountOut(firstAmountOut, tokenDir ? Y_B : X_B, tokenDir ? X_B : Y_B);

        require(secondAmountOut > amountIn, "Not profitable");

        bytes memory data = abi.encode(pairA, pairB, firstToken, secondToken, amountIn, secondAmountOut);
        pairA.swap(tokenDir ? 0 : firstAmountOut, tokenDir ? firstAmountOut : 0, address(this), data);
        uint256 profit = secondAmountOut - amountIn;
        firstToken.transfer(msg.sender, profit);
        
        return profit;
    }

    function uniswapV2Call(address sender, uint256 amount0, uint256 amount1, bytes memory data)
        public
    {
        (address pairA, address pairB, address firstToken, address secondToken, uint256 amountIn, uint256 secondAmountOut) = abi.decode(data, (address, address, address, address, uint256, uint256));

        bool tokenDir = amount0 == 0;
        IERC20(secondToken).transfer(pairB, tokenDir ? amount1 : amount0);
        IUniswapV2Pair(pairB).swap(tokenDir ? secondAmountOut : 0, tokenDir ? 0 : secondAmountOut, sender, new bytes(0));
        IERC20(firstToken).transfer(pairA, amountIn);
    }

    fallback() external {
        (
         address sender,
         uint256 amount0,
         uint256 amount1,
         bytes memory data
        ) = abi.decode(msg.data[4:], (address, uint256, uint256, bytes));

        uniswapV2Call(sender, amount0, amount1, data);
    }
}
